/////////////////////////////////////////////////////////////////////////////
// yubnub.cpp
//
// Written by Olan Patrick Barnes (patrick a-t opbarnes d-o-t com)
// Copyright (C) 2005-2006 Olan Patrick Barnes.
//
// This software is provided "as is", without any express or implied
// warranty.  In no event will the authors be held liable for any damages
// arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must not
//    claim that you wrote the original software. If you use this software
//    in a product, an acknowledgment in the product documentation would be
//    appreciated but is not required.
// 2. Altered source versions must be plainly marked as such, and must not be
//    misrepresented as being the original software.
// 3. This notice may not be removed or altered from any source distribution.
//
// Beware of bugs.
//

#include <iostream>
#include <iomanip>
#include <tchar.h>
#include <windows.h>
#include <wininet.h>
#define _ATL_CSTRING_EXPLICIT_CONSTRUCTORS  // some CString constructors will be explicit

#include <atlbase.h>
#include <atlstr.h>
#include <atlutil.h>

using namespace std;

#ifdef _UNICODE
wostream& ynt_cerr = std::wcerr;
wostream& ynt_cout = std::wcout;
#else
ostream& ynt_cerr = std::cerr;
ostream& ynt_cout = std::cout;
#endif //_UNICODE

const size_t c_sizBuffLen = 1024;

struct HtmSpecCharsT
{
	LPCTSTR szElmt;
	size_t  sizElmtLen;
	LPCTSTR szReplacement;
	bool bIsSpace;
};

HtmSpecCharsT g_arrSpecChars[] = {
	{ _T("&amp;"),   5, _T("&"),   false },
	{ _T("&copy;"),  6, _T("(C)"), false },
	{ _T("&euro;"),  6, _T("EUR"), false },
	{ _T("&gt;"),    4, _T(">"),   false },
	{ _T("&laquo;"), 7, _T("<<"),  false },
	{ _T("&ldquo;"), 7, _T("\""),  false },
	{ _T("&lsquo;"), 7, _T("'"),   false },
	{ _T("&lt;"),    4, _T("<"),   false },
	{ _T("&nbsp;"),  6, _T(" "),   true  },
	{ _T("&quot;"),  6, _T("\""),  false },
	{ _T("&raquo;"), 7, _T(">>"),  false },
	{ _T("&rdquo;"), 7, _T("\""),  false },
	{ _T("&rsquo;"), 7, _T("'"),   false }
};

size_t IsLeftSpecialHtmChar(LPCTSTR szLine, bool& bIsSpace, CString& strReplace)
{
	strReplace = _T("");
	bIsSpace = false;

	size_t sizCnt(sizeof(g_arrSpecChars) / sizeof(HtmSpecCharsT));
	for (size_t sizLp = 0; sizLp < sizCnt; sizLp++)
	{
		if (_tcsncmp(szLine,
			g_arrSpecChars[sizLp].szElmt,
			g_arrSpecChars[sizLp].sizElmtLen) == 0)
		{
			bIsSpace = g_arrSpecChars[sizLp].bIsSpace;
			strReplace = g_arrSpecChars[sizLp].szReplacement;
			return g_arrSpecChars[sizLp].sizElmtLen;
		}
	}
	return 0;
}

inline bool IsSpace_(_TINT c)
{
	switch ((char)c)
	{
	case _T(' '): case _T('\t'): case _T('\r'): case _T('\n'):
		return true;
	default:
		break;
	}
	return false;
}

void Wrap(CString& strText, int& nCurrentColumn, int nMaxColumn)
{
	if (nCurrentColumn >= nMaxColumn)
	{
		int nLoc = strText.ReverseFind(_T(' '));
		if (nLoc == -1)
		{
			return;
		}
		nCurrentColumn = strText.GetLength() - nLoc;
		strText.SetAt(nLoc, _T('\n'));
	}
}

void DeHtmlize(LPCTSTR szHtml, CString& strPlain)
{
	size_t sizSpecLen(0);
	const int nMaxCol = 80;

	LPCTSTR pchBuf;
	TCHAR szInBracs[8];
	int nInBracsStrLoc(0);
	int nLeftBracCnt(0);
	int nScriptSectionCnt(0);
	bool bSpaceAllowed(false);
	bool bNewLineAllowed(false);
	int nColumnLoc(0);

	memset(szInBracs, 0, sizeof(TCHAR) * 8);

	CStringA strCopywCharA((char)0xA9);
	CString strCopywChar(strCopywCharA);
	TCHAR chCopywChar = strCopywChar[0];

	if (szHtml != NULL)
	{
		bool bSpecIsSpace(false);
		CString strSpecRepl;

		pchBuf = szHtml;
		while (*pchBuf != _T('\0'))
		{
			if ((*pchBuf) == _T('>'))
			{
				if (nLeftBracCnt > 0)
				{
					nLeftBracCnt--;
				}

				LPTSTR szInBracsTmp = szInBracs;
				if (szInBracs[0] != _T('\0'))
				{
					::StrTrim(szInBracsTmp, _T(" \t"));
				}

				if ((_tcsncicmp(szInBracsTmp, _T("SCRIPT"), 6) == 0) ||
				    (_tcsncicmp(szInBracsTmp, _T("STYLE"), 5) == 0))
				{
					nScriptSectionCnt++;
				}
				else if ((_tcsncicmp(szInBracsTmp, _T("/SCRIPT"), 7) == 0) ||
				         (_tcsncicmp(szInBracsTmp, _T("/STYLE"), 6) == 0))
				{
					if (nScriptSectionCnt > 0)
					{
						nScriptSectionCnt--;
					}
				}
				else if ((_tcsicmp(szInBracsTmp, _T("P")) == 0)       ||
				         (_tcsncicmp(szInBracsTmp, _T("P "), 2) == 0) ||
				         (_tcsicmp(szInBracsTmp, _T("/P")) == 0)      ||
				         (_tcsicmp(szInBracsTmp, _T("/TR")) == 0)     ||
				         (_tcsicmp(szInBracsTmp, _T("/TD")) == 0))
				{
					if (bNewLineAllowed)
					{
						Wrap(strPlain, nColumnLoc, nMaxCol);
						strPlain += _T('\n');
						strPlain += _T('\n');
						nColumnLoc = 0;
						bSpaceAllowed = false;
						bNewLineAllowed = false;
					}
				}
				else if ((_tcsicmp(szInBracsTmp, _T("BR")) == 0)        ||
				         (_tcsicmp(szInBracsTmp, _T("BR /")) == 0)      ||
				         (_tcsicmp(szInBracsTmp, _T("TR")) == 0)        ||
				         (_tcsicmp(szInBracsTmp, _T("TD")) == 0)        ||
				         (_tcsncicmp(szInBracsTmp, _T("TR "), 3) == 0)  ||
				         (_tcsncicmp(szInBracsTmp, _T("TD "), 3) == 0))
				{
					if (bNewLineAllowed)
					{
						Wrap(strPlain, nColumnLoc, nMaxCol);
						strPlain += _T('\n');
						nColumnLoc = 0;
						bSpaceAllowed = false;
					}
				}
				else if ((bSpaceAllowed) &&
				         (_tcsicmp(szInBracsTmp, _T("/A")) != 0))
				{
					Wrap(strPlain, nColumnLoc, nMaxCol);
					if ( (nColumnLoc < (nMaxCol - 1)) && (nColumnLoc > 0) )
					{
						strPlain += _T(' ');
						nColumnLoc++;
						bSpaceAllowed = false;
					}
					else
					{
						strPlain += _T('\n');
						nColumnLoc = 0;
						bSpaceAllowed = false;
					}
				}

				memset(szInBracs, 0, sizeof(TCHAR) * 8);
				nInBracsStrLoc = 0;
			}
			else if ((*pchBuf) == _T('<'))
			{
				nLeftBracCnt++;
				memset(szInBracs, 0, sizeof(TCHAR) * 8);
				nInBracsStrLoc = 0;
			}
			else if (nLeftBracCnt > 0)
			{
				if (!IsSpace_(*pchBuf))
				{
					if (nInBracsStrLoc < 7)
					{
						szInBracs[nInBracsStrLoc++] = *pchBuf;
					}
				}
			}
			else if (nScriptSectionCnt == 0)
			{
				bSpecIsSpace = false;
				CString strSpecRepl;

				if (*pchBuf == chCopywChar)
				{
					sizSpecLen = 1;
					strSpecRepl = _T("(C)");
				}
				else
				{
					sizSpecLen = IsLeftSpecialHtmChar(pchBuf,
									bSpecIsSpace,
									strSpecRepl);
				}

				if (sizSpecLen > 0)
				{
					if (bSpecIsSpace)
					{
						if (bSpaceAllowed)
						{
							Wrap(strPlain, nColumnLoc, nMaxCol);
							strPlain += strSpecRepl;
							nColumnLoc += strSpecRepl.GetLength();
							bSpaceAllowed = false;
						}
					}
					else
					{
						strPlain += strSpecRepl;
						nColumnLoc += strSpecRepl.GetLength();
						bSpaceAllowed = true;
						bNewLineAllowed = true;
						Wrap(strPlain, nColumnLoc, nMaxCol);
					}
					pchBuf = _tcsninc(pchBuf, sizSpecLen - 1);
				}
				else if (!IsSpace_(*pchBuf))
				{
					strPlain += *pchBuf;
					nColumnLoc++;
					bSpaceAllowed = true;
					bNewLineAllowed = true;
					Wrap(strPlain, nColumnLoc, nMaxCol);
				}
				else if (bSpaceAllowed)
				{
					Wrap(strPlain, nColumnLoc, nMaxCol);
					if ( (nColumnLoc < (nMaxCol - 1)) && (nColumnLoc > 0) )
					{
						strPlain += _T(' ');
						nColumnLoc++;
						bSpaceAllowed = false;
					}
					else
					{
						strPlain += _T('\n');
						nColumnLoc = 0;
						bSpaceAllowed = false;
					}
				}
			}
			pchBuf = _tcsinc(pchBuf);
		}
	}
}

bool GetPlainTextFromUrl(LPCTSTR szUrl, CString& strPlain, bool bRaw)
{
	bool bCreateOk(true);
	CString strUrlOutput;

	HINTERNET hSes;
	hSes = InternetOpen(_T("yubnubopb"),
		INTERNET_OPEN_TYPE_PRECONFIG,
		NULL,
		NULL,
		0);

	HINTERNET hNew = NULL;
	if (hSes == NULL)
	{
		bCreateOk = false;
	}
	else
	{
		hNew = InternetOpenUrl(hSes, szUrl, NULL, 0, INTERNET_FLAG_DONT_CACHE, 0);
	}

	if (hNew == NULL)
	{
		bCreateOk = false;
	}
	else
	{
		TCHAR szBuffer[c_sizBuffLen];
		DWORD dwRead(0);

		// read until dwRead == 0.  InternetReadFile will not actually return
		// a FALSE condition upon the file end.
		while (::InternetReadFile(hNew,
			szBuffer,
			c_sizBuffLen - 1,
			&dwRead))
		{
			if (dwRead == 0)
			{
				break;
			}

			// Null terminate temp buffer.
			szBuffer[dwRead] = _T('\0');
			strUrlOutput += szBuffer;
		}
	}

	BOOL bCloseOk(FALSE);
	if (hNew != NULL)
	{
		bCloseOk = InternetCloseHandle(hNew);
		hNew = NULL;
	}

	if (hSes != NULL)
	{
		if (!InternetCloseHandle(hSes))
		{
			bCloseOk = FALSE;
		}
		else
		{
			hSes = NULL;
		}
	}

	if (bRaw)
	{
		strPlain = strUrlOutput;
	}
	else
	{
		DeHtmlize((LPCTSTR)strUrlOutput, strPlain);
	}
	return (bCreateOk && bCloseOk);
}

int _tmain(int argc, TCHAR** argv)
{
	int nRet(0);

	ynt_cout << _T("yubnub command line tool ver. 1.1")
		<< _T(" - Copyright (C) 2005-2006 Olan Patrick Barnes") << endl;

	CString strUriLine = _T("http://www.yubnub.org/parser/parse?command=");

	bool bDoHelp(false);
	if (argc > 1)
	{
		if ( (_tcsicmp(argv[1], _T("-h")) == 0)     ||
			(_tcsicmp(argv[1], _T("-?")) == 0)     ||
			(_tcsicmp(argv[1], _T("-help")) == 0)  ||
			(_tcsicmp(argv[1], _T("--help")) == 0) ||
			(_tcsicmp(argv[1], _T("/h")) == 0)     ||
			(_tcsicmp(argv[1], _T("/?")) == 0)     ||
			(_tcsicmp(argv[1], _T("/help")) == 0))
		{
			bDoHelp = true;
			argc = 1;
		}
	}

	if (argc == 1)
	{
		if (!bDoHelp)
		{
			ynt_cerr << endl
				<< "Nothing to do.  Type \"yubnub --help\" for help."
				<< endl;
		}
		else
		{
			ynt_cout << endl
				<< _T("USAGE: yubnub [options] [yubnub command line]")
				<< endl
				<< endl
				<< _T("OPTIONS:") << endl
				<< _T("   --help    - help") << endl
				<< _T("   --browser - launch browser to run command (default).") << endl
				<< _T("   --plain   - output results to console in plain text.") << endl
				<< _T("   --raw     - output the raw html results to the console.") << endl
				<< endl
				<< endl
				<< _T("See http://www.yubnub.org/ for more information.") << endl
				<< endl
				<< endl;
		}
	}
	else
	{
		bool bLaunchBrowser(true);
		bool bExplicitLaunchBrowser(false);
		bool bShowPlain(false);
		bool bShowRaw(false);

		for (int nLp = 1; nLp < argc; nLp++)
		{
			if (_tcsicmp(argv[nLp], _T("--raw")) == 0)
			{
				bShowRaw = true;
				if (!bExplicitLaunchBrowser)
				{
					bLaunchBrowser = false;
				}
				continue;
			}

			if (_tcsicmp(argv[nLp], _T("--plain")) == 0)
			{
				bShowPlain = true;
				if (!bExplicitLaunchBrowser)
				{
					bLaunchBrowser = false;
				}
				continue;
			}

			if (_tcsicmp(argv[nLp], _T("--browser")) == 0)
			{
				bLaunchBrowser = true;
				bExplicitLaunchBrowser = true;
				continue;
			}

			CStringW strUriParam(argv[nLp]);
			DWORD dwResLen(0);
			DWORD dwBufLen((strUriParam.GetLength() * 4) + 1);
			CStringW strUriParamEscaped;
			LPWSTR szEsc = strUriParamEscaped.GetBuffer(dwBufLen);
			AtlEscapeUrl((LPCWSTR)strUriParam,
				szEsc,
				&dwResLen,
				dwBufLen,
				ATL_URL_NO_META | ATL_URL_ENCODE_PERCENT);
			if (dwResLen < dwBufLen)
			{
				strUriParamEscaped.ReleaseBuffer(dwResLen);
			}
			else
			{
				strUriParamEscaped.ReleaseBuffer(0);
				dwBufLen = dwResLen + 1;
				szEsc = strUriParamEscaped.GetBuffer(dwBufLen);
				AtlEscapeUrl((LPCWSTR)strUriParam,
					szEsc,
					&dwResLen,
					dwBufLen,
					ATL_URL_NO_META | ATL_URL_ENCODE_PERCENT);
				if (dwResLen < dwBufLen)
				{
					strUriParamEscaped.ReleaseBuffer(dwResLen);
				}
				else
				{
					strUriParamEscaped.ReleaseBuffer(0);
				}
			}

			strUriLine += strUriParamEscaped;
			if (nLp < (argc - 1))
			{
				strUriLine += _T("+");
			}
		}

		if (bShowPlain || bShowRaw)
		{
			CString strOut;
			GetPlainTextFromUrl((LPCTSTR)strUriLine, strOut, bShowRaw);
			ynt_cout << strOut;
		}

		if (bLaunchBrowser)
		{
			ynt_cout << endl << _T("Launching: ") << (LPCTSTR)strUriLine << endl;

			SHELLEXECUTEINFO sei;
			ZeroMemory(&sei, sizeof(SHELLEXECUTEINFO));
			sei.cbSize = sizeof(SHELLEXECUTEINFO);
			sei.lpFile = (LPCTSTR)strUriLine;
			sei.nShow = SW_SHOW;
			sei.fMask = SEE_MASK_DOENVSUBST |
				SEE_MASK_INVOKEIDLIST |
				SEE_MASK_FLAG_NO_UI |
				SEE_MASK_FLAG_DDEWAIT;
			sei.lpVerb = NULL;
			BOOL bSucceeded = ShellExecuteEx(&sei);
			if (!bSucceeded)
			{
				ynt_cerr << _T("Error - Launching webpage failed: ")
					<< endl << _T("   ") << (LPCTSTR)strUriLine
					<< endl;
				nRet = 1;
			}
		}
	}
	return nRet;
}
